/*
** Module   : EventList.java
** Abstract : encodes events and widgets that determine the condition of
**            termination of the Progress WAIT-FOR statement
**
** Copyright (c) 2007-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20070115   @31883 Created based on the WaitList.java file.
**                           Changed the class name and redesigned the
**                           implementation to allow triggers and exit
**                           conditions to coexist.
** 002 NVS 20070417   @33087 lookup() method was changed to meet the newly
**                           found requirements for extending the scope of
**                           some triggers (universal events) from widget
**                           to frame and all exit events from widgets to
**                           frame to session window.
** 003 SIY 20070706   @34391 Added chaining of event lists.
** 004 NVS 20071107   @35726 Fixed bug in Go-On events handling. They were
**                           stored in the original case causing name
**                           mismatch and lookup failure.
** 005 GES 20080314   @37449 Added working ANY-PRINTABLE and a disabled
**                           prototype of ANY-KEY support.
** 006 NVS 20080320   @37617 Added anyEventForWidget() lookup method.
** 007 NVS 20080625   @38919 Extended event lookup for session window to
**                           match anywhere events.
** 008 SIY 20080709   @39082 ENTER and RETURN should be handled as
**                           aliases when checking for GO events.
**                           Inferred generics.
** 009 SIY 20080901   @39638 Added ANY-KEY event support.
** 010 SIY 20080903   @39696 Fixed order of checks in ANY-KEY support.
** 011 CA  20081204   @40816 Added getRowDisplayEvents, which collects 
**                           all ROW-DISPLAY events.
** 012 CA  20090408   @41733 Added getExitWidgetIds - returns widget ids
**                           registered for exit events.
** 013 GES 20090629   @42969 Added isSame() helper to test equality. Improved
**                           serialization efficiency. 
** 014 SIY 20101122          Minor cleanups in order to suppress (incorrect)
**                           warning.
** 015 SIY 20111101          Updated imports.
** 016 CS  20130204          implement conversion support for sockets - added
**                           addEvent(String, handle) to implement events for
**                           non-widgets.
*/

package com.goldencode.p2j.ui;

import java.io.*;
import java.util.*;

import com.goldencode.p2j.util.Utils;
import com.goldencode.p2j.util.handle;

/**
 * This class supports the lists of events and widgets that define the
 * termination of implicit or explicit WAIT-FOR statement and triggers 
 * activation.
 * <p>
 * ANYWHERE option is kept either globally, or per widget list, but they can't
 * be mixed.
 * <p>
 * Events are kept as Sets of symbolic names. Widgets are kept as integer IDs. 
 * <p>
 */
public class EventList
implements Externalizable
{
   /** Low level WAIT-FOR operation code */
   public static final int WAIT_FOR = 0;

   /** High level PROMPT-FOR operation code */
   public static final int PROMPT_FOR = 1;

   /** High level SET operation code */
   public static final int SET = 2;

   /** High level UPDATE operation code */
   public static final int UPDATE = 3;

   /** High level CHOOSE operation code */
   public static final int CHOOSE = 4;

   /** The lowest code */
   static final int CODE_MIN = WAIT_FOR;

   /** The highest code */
   static final int CODE_MAX = CHOOSE;

   /** Type of operation */
   private int operation = WAIT_FOR;

   /** Holds the GO-ON keys */
   private List<String> goOn = new LinkedList<String>();

   /** Set of validatable widgets */
   private Set<Integer> validations = new HashSet<Integer>();

   /** List of event definitions */
   private List<EventDefinition> events = new ArrayList<EventDefinition>();

   /** Chain of event lists (not transferred via network).  */
   private EventList chain = null;
   
   /**
    * Default constructor
    */
   public EventList()
   {
   }

   /**
    * A constructor that adds some default events.
    *
    * @param   flag
    *          does nothing but simply selects the constructor
    */
   public EventList(boolean flag)
   {
      addEvent("GO");
      addGoOn("GO");
   }

   /**              
    * Reports if the list of event definitions is the same as the current
    * instance. The GO-ON list, validations list and any chaining is ignored.
    *
    * @param    other
    *           The event list to compare against.
    *
    * @return   <code>true</code> if the event definitions are the same.
    */
   public boolean isSame(EventList other)
   {
      int len = events.size();
      
      if (len == other.events.size())
      {
         for (int i = 0; i < len; i++)
         {
            EventDefinition ed1 = events.get(i);
            EventDefinition ed2 = other.events.get(i);
            
            if (!ed1.isSame(ed2))
            {
               return false;
            }
         }
         
         return true;
      }
      
      return false;
   }
   
   /**
    * Set chain of other event lists.
    * 
    * @param   chain
    *          The chain of event lists.
    */
   public void setChain(EventList chain)
   {
      this.chain = chain;
   }

   /**
    * Search for registered ROW-DISPLAY events.
    * 
    * @return   a map containing pairs of (widgetId, triggerId). May be null.
    */
   public Map<Integer, Integer> getRowDisplayEvents()
   {
      Map<Integer, Integer> res = null;
      
      for (EventDefinition ed : events)
      {
         Map<Integer, Integer> m = ed.getRowDisplayEvents();
         if (m != null)
         {
            if (res == null)
            {
               res = new HashMap<Integer, Integer>();
            }
            res.putAll(m);
         }
      }
      
      return res;
   }
   
   /**
    * Get the widget id list for which there are defined exit events.
    * 
    * @return   the widget id list, or <code>null</code> if there is no exit 
    *           event defined.
    */
   public int[] getExitWidgetIds()
   {
      Set<Integer> res = null;
      
      for (EventDefinition ed : events)
      {
         if (ed.getTriggerId() == -1)
         {
            // this is an exit event, add registered widgets to the list
            Set<Integer> w = ed.getWidgetIds();
            if (w != null)
            {
               if (res == null)
               {
                  res = new HashSet<Integer>(3);
               }

               res.addAll(w);
            }
         }
      }
      
      return res == null ? null : Utils.integerCollectionToPrimitive(res);
   }

   /**
    * Adds an event definition to the list.
    * If ANYWHERE keyword was used, the appropriate flag is set for this
    * entry.
    * <p>
    * Event names are case-insensitive on input and stored uppercased.
    *
    * @param event
    *        array of string event definitions
    * @param widget
    *        array of <code>GenericWidget</code>instances associated with
    *        the specified events
    * @param anywhere
    *        <code>true</code> if ANYWHERE option is used for this entry.
    */
   public void addEvent(String[]        event,
                        GenericWidget[] widget,
                        boolean         anywhere)
   {
      // sanity checks
      if (event == null  || event.length == 0 || 
         (widget != null && widget.length == 0))
         return;

      int i = 0;

      // create a set of events
      Set<String> s = new HashSet<String>();
      String caseEvent = null;
      for (i = 0; i < event.length; i ++)
      {
         caseEvent = setCase(event[i]);
         
         // alternate key labels won't be matched if we don't convert them
         // to the primary label
         
         // convert the label to a key code, alternate labels are honored
         // here 
         int code = Keyboard.keyCode(caseEvent);
         
         // is this a valid label?
         if (code != -1)
         {
            // yes, it was a label so now convert the common code to the
            // primary label name
            String primary = Keyboard.keyLabel(code); 
            caseEvent = (primary.length() > 0) ? primary : caseEvent;
         }
         
         s.add(caseEvent);
      }

      // create a set of widget IDs
      Set<Integer> w = null;
      if (widget != null)
      {
         w = new HashSet<Integer>();
         for (i = 0; i < widget.length; i ++)
         {
            w.add(new Integer(widget[i].getId()));
         }
      }
      else
         // do not set specific anywhere flag if the event is global anywhere
         anywhere = false;
      
      // add new event definition to the list
      events.add(new EventDefinition(s, w, anywhere));
   }

   /**
    * Adds an event definition to the list without ANYWHERE option.
    * Event names are case-insensitive on input and stored uppercased.
    *
    * @param event
    *        array of event names
    * @param widget
    *        array of <code>GenericWidget</code>instances associated with
    *        the specified events
    */
   public void addEvent(String[] event, GenericWidget[] widget)
   {
      addEvent(event, widget, false);
   }

   /**
    * Adds an event definition to the list without ANYWHERE option.
    *
    * @param event
    *        event name
    * @param widget
    *        array of <code>GenericWidget</code>instances associated with
    *        the specified event
    */
   public void addEvent(String event, GenericWidget[] widget)
   {
      addEvent(new String[] {event}, widget, false);
   }

   /**
    * Adds an event definition to the list without ANYWHERE option.
    *
    * @param event
    *        array of event names
    * @param widget
    *        <code>GenericWidget</code>instance associated with
    *        the specified events
    */
   public void addEvent(String[] event, GenericWidget widget)
   {
      addEvent(event, new GenericWidget[] {widget}, false);
   }

   /**
    * Adds an event definition to the list without ANYWHERE option.
    *
    * @param event
    *        event name
    * @param widget
    *        <code>GenericWidget</code>instance associated with
    *        the specified event
    */
   public void addEvent(String event, GenericWidget widget)
   {
      addEvent(new String[] {event}, new GenericWidget[] {widget}, false);
   }

   /**
    * Adds an event definition to the list.
    * If ANYWHERE keyword was used, the appropriate flag is set for this
    * entry.
    *
    * @param event
    *        array of event names
    * @param widget
    *        <code>GenericWidget</code>instance associated with
    *        the specified events
    * @param anywhere
    *        <code>true</code> if ANYWHERE option is used for this entry.
    */
   public void addEvent(String[]      event,
                        GenericWidget widget,
                        boolean       anywhere)
   {
      addEvent(event, new GenericWidget[] {widget}, anywhere);
   }

   /**
    * Adds an event definition to the list.
    * If ANYWHERE keyword was used, the appropriate flag is set for this
    * entry.
    *
    * @param event
    *        event name
    * @param widget
    *        array of <code>GenericWidget</code>instances associated with
    *        the specified event
    * @param anywhere
    *        <code>true</code> if ANYWHERE option is used for this entry.
    */
   public void addEvent(String          event, 
                        GenericWidget[] widget,
                        boolean         anywhere)
   {
      addEvent(new String[] {event}, widget, anywhere);
   }

   /**
    * Adds an event definition to the list.
    * If ANYWHERE keyword was used, the appropriate flag is set for this
    * entry.
    *
    * @param event
    *        event name
    * @param widget
    *        <code>GenericWidget</code>instance associated with
    *        the specified event
    * @param anywhere
    *        <code>true</code> if ANYWHERE option is used for this entry.
    */
   public void addEvent(String        event,
                        GenericWidget widget,
                        boolean       anywhere)
   {
      addEvent(new String[] {event},
               new GenericWidget[] {widget},
               anywhere);
   }
   
   /**
    * Adds an event definition to the list.
    * If ANYWHERE keyword was used, the appropriate flag is set for this
    * entry.
    *
    * @param event
    *        event name
    * @param handle
    *        The handle associated with the specified event.
    *
    */
   public void addEvent(String event, handle handle)
   {
      if(handle.get() instanceof CommonFrame)
      {
         addEvent(event, ((CommonFrame) handle.get()).asWidget());
      }
      else
      {
         // TODO implement addEvent for sockets
      }
      
   }

   /**
    * Adds an event definition to the list without ANYWHERE option.
    *
    * @param event
    *        event name
    * @param frame
    *        <code>CommonFrame</code> instance associated with the specified
    *        event.
    */
   public void addEvent(String event, CommonFrame frame)
   {
      addEvent(new String[] {event}, frame);
   }
   
   /**
    * Adds an event definition to the list without ANYWHERE option.
    *
    * @param event
    *        array of event names
    * @param frame
    *        <code>CommonFrame</code> instance associated with the specified
    *        event.
    */
   public void addEvent(String[] event, CommonFrame frame)
   {
      GenericWidget widget = frame.asWidget();
      addEvent(event, new GenericWidget[] {widget}, false);
   }
   
   /**
    * Adds an event definition to the list.
    *
    * @param event
    *        array of event names
    */
   public void addEvent(String[] event)
   {
      addEvent(event, (GenericWidget[])null, false);
   }

   /**
    * Adds an event definition to the list.
    *
    * @param event
    *        array of event names
    * @param anywhere
    *        <code>true</code> if the global ANYWHERE option is being set.
    */
   public void addEvent(String[] event, boolean anywhere)
   {
      addEvent(event, (GenericWidget[])null, false);
   }

   /**
    * Adds an event definition to the list.
    *
    * @param event
    *        event name
    */
   public void addEvent(String event)
   {
      addEvent(new String[] {event}, (GenericWidget[])null, false);
   }

   /**
    * Adds an event definition to the list.
    *
    * @param event
    *        event name
    * @param anywhere
    *        <code>true</code> if the global ANYWHERE option is being set.
    */
   public void addEvent(String event, boolean anywhere)
   {
      addEvent(new String[] {event}, (GenericWidget[])null, false);
   }

   /**
    * Adds a key to the list of GO-ON keys.
    *
    * @param key
    *        array of key labels
    */
   public void addGoOn(String[] key)
   {
      for (int i = 0; i < key.length; i ++)
         goOn.add(setCase(key[i]));
   }

   /**
    * Adds a key to the list of GO-ON keys.
    *
    * @param key
    *        key label
    */
   public void addGoOn(String key)
   {
      goOn.add(setCase(key));
   }

   /**
    * Sets the registered trigger for this event list.
    *
    * @param triggerId
    *        trigger ID
    */
   public void setTriggerId(int triggerId)
   {
      for (EventDefinition ed : events)
      {
         ed.setTriggerId(triggerId);
      }
   }

   /**
    * Gets trigger ID off of this event list.
    *
    * @return  an ID of a trigger or <code>-1</code> if this event list
    *          is for exit conditions.
    */
   public int getTriggerId()
   {
      if (events.size() == 0)
         return -1;
      
      return events.get(0).getTriggerId();
   }

   /**
    * Check if event is listed in GO-ON array.
    * 
    * @param    event
    *           Event name.
    *
    * @return   <code>true</code> if event is mentioned in the GO-ON array.
    */
   public boolean isGoEvent(String event)
   {
      // convert the event name to the proper case
      String ev = setCase(event);
      
      for (String go : goOn)
      {
         if(go.equals(ev))
            return true;
      }

      // RETURN and ENTER are aliases
      // Note that isGoEvent always receives "RETURN"
      // as event name, because it is obtained by
      // conversion of key code into event name.
      if ("RETURN".equals(ev))
      {
         ev = "ENTER";
         
         for (String go : goOn)
         {
            if(go.equals(ev))
               return true;
         }
      }
      
      return (chain == null) ? false : chain.isGoEvent(event);
   }

   /**
    * Sets the validateable widgets.
    * 
    * @param  widget
    *         <code>GenericWidget</code> instance
    */
   public void setValidatable(GenericWidget widget)
   {
      validations.add(new Integer(widget.getId()));
   }

   /**
    * Check if a given widget is validateable.
    * 
    * @param  widget
    *         <code>GenericWidget</code> instance
    * @return  <code>true</code> if the widget is validateable.
    */
   public boolean isValidatable(GenericWidget widget)
   {
      return validations.contains(new Integer(widget.getId()));
   }

   /**
    * Check if a given widget is validateable.
    * 
    * @param    widgetId
    *           widget ID
    *
    * @return   <code>true</code> if the widget is validatable.
    */
   public boolean isValidatable(int widgetId)
   {
      return validations.contains(new Integer(widgetId));
   }

   /**
    * Queries the terminal I/O operation type set here.
    *
    * @return operation type code
    */
   public int getOperation()
   {
      return operation;
   }

   /**
    * Sets the terminal I/O operation type code for this instance.
    * The operation code can only be set once while the current operation code
    * is set to WAIT_FOR, which is the default. Subsequent requests are simply
    * ignored.
    *
    * @param code
    *        operation type code as one of the constants defined above.
    */
   public void setOperation(int code)
   {
      if (operation != WAIT_FOR)
         return;
      
      if (code < CODE_MIN || code > CODE_MAX)
         return;

      operation = code;
   }

   /**
    * Merges in all events for all widgets and GO-ON events from the specified
    * <code>EventList</code>
    *
    * @param from
    *        another <code>EventList</code> to merge with <code>this</code>.
    * @param lifo
    *        <code>true</code> if the events are imported in LIFO mode.
    */
   public void merge(EventList from, boolean lifo)
   {
      // copy exit conditions
      if (lifo)
         events.addAll(0, from.events);
      else
         events.addAll(from.events);

      // copy GO-ON data
      goOn.addAll(from.goOn);
   }

   /**
    * Checks whether this event matches any event defined in this
    * <code>EventList</code> instance.
    * <p>
    * The search is done slightly differently for the triggers versus exit
    * events and for various classes of events.
    * <p>
    * According to the known facts, there are at least two classes of events:
    * <ul>
    *  <li>universal events
    *  <li>regular events
    * </ul>
    * Universal events have the ability to extend the search for triggers,
    * whereas regular events don't. Both classes of events have the ability
    * to extend the search for exit conditions.
    * <p>
    * The search for triggers is done in the following order:
    * <ul>
    *  <li>a match for the specified widget and event;
    *  <li>if anywhere option is specified, a match for the specified event
    *       anywhere within the frame;
    *  <li>if it is a universal event, a match for the specified event and
    *      the containing frame;
    *  <li>if global anywhere option is specified, a match for the event
    *      anywhere globally.
    * </ul>
    * <p>
    * If the widget is a frame or the session window, only the first step is
    * taken.
    *<p>
    * The search for exit conditions is done in the following order:
    * <ul>
    *  <li>a match for the specified widget and event;
    *  <li>if anywhere option is specified, a match for the specified event
    *       anywhere within the frame;
    *  <li>a match for the specified event and the containing frame;
    *  <li>if global anywhere option is specified, a match for the event
    *      anywhere globally.
    *  <li>a match for the specified event and the session window.
    * </ul>
    * <p>
    * If the widget is a frame, the search starts form the step 3.
    * <br>
    * If the widget is the session window, only the first step is taken.
    *
    * @param    frameId
    *           Frame ID, can be <code>-1</code>
    * @param    widgetId
    *           widget ID, can't be <code>-1</code>, but can be 
    *           <code>SESSION_WINDOW_ID</code>
    * @param    eventCode
    *           numeric event code
    * @param    trigger
    *           flag, telling the search is for a trigger (<code>true</code>)
    *           or for an exit condition (<code>false</code>).
    * @param    isKey
    *           Marks real key pressed by user.
    *           
    * @return   returns trigger ID or -1 for triggers;
    *           returns event index or -1 for exit conditions. In either case,
    *           -1 means no match was found.
    */
   public int lookup(int     frameId,
                     int     widgetId,
                     int     eventCode,
                     boolean trigger, 
                     boolean isKey)
   {
      int result = lookupWorker(frameId, widgetId, eventCode, trigger, isKey);
      
      return (result == -1 && chain != null) 
             ? chain.lookup(frameId, widgetId, eventCode, trigger, isKey)
             : result;
   }
   
   /**
    * Checks whether this event list has common events and widgets with
    * the specified one.
    * <p>
    * They intersect if any of the event definitions in this list intersects
    * with any event definition from the given list.
    *
    * @param    el
    *           another instance of <code>EventList</code>
    * 
    * @return   <code>true</code> if the event lists have common events and
    *           widgets
    */
   public boolean intersects(EventList el)
   {
      for (EventDefinition ed : events)
      {
         for (EventDefinition oed : el.events)
         {
            if (ed.intersects(oed))
               return true;
         }
      }
      
      return false;
   }
   
   /**
    * Returns an instance of <code>EventList</code> which represents
    * the common events and widgets with the specified one.
    * <p>
    * The returned <code>EventList</code> retains all the internal fields of
    * this one except the list of event definitions.
    *
    * @param    el
    *           another instance of <code>EventList</code>
    * 
    * @return   common part as a <code>EventList</code> or <code>null</code>
    *           if there is nothing in common.
    */
   public EventList intersection(EventList el)
   {
      LinkedList<EventDefinition> l = new LinkedList<EventDefinition>();

      for (EventDefinition ed : events)
      {
         for (EventDefinition oed : el.events)
         {
            if (!ed.intersects(oed))
               continue;
            
            l.add(ed.intersection(oed));
         }
      }
      
      // create the resulting event list
      if (l.size() == 0)
         return null;
      
      EventList nel   = new EventList();
      nel.operation   = this.operation;
      nel.goOn        = this.goOn;
      nel.validations = this.validations;
      nel.events      = l;
      return nel;
   }
   
   /**
    * Returns a list of zero or more instances of <code>EventList</code>s
    * which represent the events and widgets of this instance which are not
    * in the specified one.
    *
    * @param    el
    *           another instance of <code>EventList</code>
    * 
    * @return   a list of <code>EventList</code> or <code>null</code>
    *           if there is nothing uncommon.
    */
   public EventList difference(EventList el)
   {
      if (!intersects(el))
         return null;

      List<EventDefinition> from = events;
      
      // this outer loop is over all event definitions to be subtracted

      for (EventDefinition oed : el.events)
      {
         LinkedList<EventDefinition> l = new LinkedList<EventDefinition>();
         
         for (EventDefinition ed : from)
         {
            if (!ed.intersects(oed))
            {
               l.add(ed);
               continue;
            }
            
            l.addAll(ed.difference(oed));
         }
         
         from = l;
      }
      
      // create the resulting event list
      EventList nel   = new EventList();
      nel.operation   = this.operation;
      nel.goOn        = this.goOn;
      nel.validations = this.validations;
      nel.events      = from;
      return nel;
   }
   
   /**
    * Checks whether this definition contains exit conditions referring
    * to the specified widget.
    *
    * @param    widgetId
    *           widget ID to check
    * 
    * @return   <code>true</code> if the widget is referred from here
    */
   public boolean anyEventForWidget(int widgetId)
   {
      for (EventDefinition ed : events)
      {
         if (ed.anyEventForWidget(widgetId))
         {
            return true;
         }
      }

      return false;
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      operation = in.readInt();
      Collections.addAll(goOn, (String[]) in.readObject()); 
      
      int[] vals = (int[]) in.readObject();
      validations.addAll(Utils.primitiveToIntegerList(vals));
      
      Collections.addAll(events, (EventDefinition[]) in.readObject());
      
      // chaining is not transferred
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeInt(operation);
      out.writeObject(goOn.toArray(new String[0]));      
      out.writeObject(Utils.integerCollectionToPrimitive(validations));
      out.writeObject(events.toArray(new EventDefinition[0]));
      
      // chaining is not transferred
   }
   
   /**
    * Conditionally converts the event name to the upper case.
    * Event names that correspond to the key labels are unchanged. The rest
    * is uppercased.
    *
    * @param  event
    *         original event name
    * @return event name suitable for storing
    */
   private String setCase(String event)
   {
      int kc = Keyboard.keyCode(event);

      if (kc == -1 || event.length() > 1) 
      {
         // translate to upper case
         return event.toUpperCase();
      }
      
      // a single character key label, return as is
      return event;
   }

   /**
    * Worker routine for {@link #lookup(int, int, int, boolean, boolean)}.
    * 
    * @param    frameId
    *           Frame ID, can be <code>-1</code>
    * @param    widgetId
    *           widget ID, can't be <code>-1</code>, but can be 
    *           <code>SESSION_WINDOW_ID</code>
    * @param    eventCode
    *           numeric event code
    * @param    trigger
    *           flag, telling the search is for a trigger (<code>true</code>)
    *           or for an exit condition (<code>false</code>).
    * @param    isKey
    *           This flags marks reals keys typed by user.
    * @return  returns trigger ID or -1 for triggers;
    *          returns event index or -1 for exit conditions. In either case,
    *          -1 means no match was found.
    */
   private int lookupWorker(int     frameId, 
                            int     widgetId, 
                            int     eventCode, 
                            boolean trigger, 
                            boolean isKey)
   {
      int result = -1;
      
      // obtain the event name and class
      String event = Keyboard.eventName(eventCode);
      boolean universal = EventDefinition.isEventUniversal(eventCode);
      
      // convert the event name to the proper case
      String ev = setCase(event);
      
      // look the list up for the specified event and widget
      for (EventDefinition ed : events)
      {
         result = ed.lookup(-1, widgetId, ev, trigger, true, false, false);
         
         if (result != -1)
            return result;
      }
      
      // honor ANY-PRINTABLE event (must be before ANY-KEY)
      if (Keyboard.isPrintable(ev))
      {
         for (EventDefinition ed : events)
         {
            result = ed.lookup(-1, widgetId, Keyboard.ANY_PRINTABLE, trigger,
                               true, false, false);
            
            if (result != -1)
               return result;
         }
      }

      // look for ANY-KEY
      if (isKey)
      {
         for (EventDefinition ed : events)
         {
            result = ed.lookup(-1, widgetId, Keyboard.ANY_KEY, trigger, 
                               true, false, true);
            
            if (result != -1)
               return result;
         }
      }

      // look the list up for the specified event anywhere in the frame
      for (EventDefinition ed : events)
      {
         result = ed.lookup(frameId, widgetId, ev, trigger, false, 
                            universal, isKey);
         if (result != -1)
            return result;
      }

      // look the list up for the specified event and the containing frame
      if ((trigger && universal) || !trigger)
      {
         for (EventDefinition ed : events)
         {
            result = ed.lookup(-1, frameId, ev, trigger, true, false, isKey);
            if (result != -1)
               return result;
         }
      }

      // look the list up for the event anywhere globally
      for (EventDefinition ed : events)
      {
         result = ed.lookup(-1, -1, ev, trigger, false, false, isKey);
         if (result != -1)
            return result;
      }

      // look the list up for the specified event and the session window
      if (!trigger)
      {
         for (EventDefinition ed : events)
         {
            result = ed.lookup(-1, WindowConfig.SESSION_WINDOW_ID, ev,
                               trigger, true, false, isKey); 
            if (result != -1)
               return result;
         }
      }

      return -1;
   }
}
